/*
   f98ip.h for f98ip.c
   FBBI's Funge-98 Instruction Pointer - header.
   v0.98 Oct 1 1998 Chris Pressey
   v0.98a Mar 26 2003 Chris Pressey

   Copyright (c)1998, 2001, 2003, Cat's Eye Technologies.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

     Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

     Neither the name of Cat's Eye Technologies nor the names of its
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission. 

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
   CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
   OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
   ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE. 

*/

#ifndef _F98IP_H_
#define _F98IP_H_

typedef struct InstructionPointer
{
  cell ir;      /* instruction register */
  cell x, y;    /* position */
  cell dx, dy;  /* delta */
  cell sx, sy;  /* offset */
  cell a, b;    /* scratch storage */
  cell ec;      /* error code */
  flag sm;      /*-- stringmode */
  flag pm;      /*-- spacemode */
  flag hm;      /*-- haltmode */
  bfspace * bs; /* befunge-space */
  void * s2;   /* stack-stack */
  void (*fingins[26][26])(struct InstructionPointer *);
  int finglev[26]; /* fingerprints */
  int ssize;
  int sssize;
} ip;

extern ip * ip_alloc(bfspace * bs, int ssize, int sssize);
extern void ip_free(ip * i);
extern void ip_reverse(ip * i);
extern void ip_march(ip * i);
extern void ip_backtrack(ip * i);
extern void ip_advance(ip * i);
extern void ip_move(ip * i);

void * ip_stack(ip * i);
cell ip_pop(ip * i);
int ip_push(ip * i, cell x);

int ip_clearstack(ip * i);
cell ip_popoffset(ip * i, int offset);
void ip_pushoffset(ip * i, int offset, cell value);
int ip_allocstack(ip * i);
void ip_freestack(ip * i);
int ip_transfer(ip * i, int val);
void ip_pop_string(ip * i, char * s);
void ip_push_string(ip * i, char * s);
int ip_stack_measure(ip * i);
int ip_stacks_measure(ip * i);
int ip_stack_measure_offset(ip * i, int offset);
cell ip_stack_peek(ip * i, int offset);
cell ip_stack_peek_peek(ip * i, int o1, int o2);

#endif /* _F98IP_H_ not defined */
